
__title__ = 'mbuslite'
__version__ = '0.0.4'
__description__ = 'A small, in-process message bus implementation.'
__author__ = 'Francis Ferrell'
__author_email__ = 'francisferrell@gmail.com'
__copyright__ = f'2020 {__author__}'
__url__ = 'https://gitlab.com/francisferrell/mbuslite'
__license__ = 'MIT'



from .bus import MessageBus
from .service import Service

Bus = MessageBus()
"""MessageBus: the default, application-wide bus"""





class Singleton( type ):
    """
    A textbook singleton implementation as a metaclass.
    """

    _instances = {}

    def __call__( cls, *args, **kwargs ):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__( *args, **kwargs )
        return cls._instances[cls]



class Service( metaclass = Singleton ):
    """
    A singleton base class.
    """





from collections import defaultdict
from typing import Callable



class Subscription:
    """
    Opaque object allowing the subscriber to unsubscribe.
    """

    def __init__( self, bus, topic, handler ):
        self.bus = bus
        self.topic = topic
        self.handler = handler


    def unsubscribe( self ):
        """
        Remove the subscription created when ``Bus.subscribe`` was called.
        """
        self.bus.subscriptions[self.topic].remove( self )



class MessageBus:
    """
    The message bus. Should be used by importing ``mbuslite.Bus``.
    """

    def __init__( self ):
        self.subscriptions = defaultdict( list )


    def publish( self, topic: str, *args, **kwargs ) -> None:
        """
        Publish a message to the subscribers of the specified ``topic``. All remaining arguments and
        keyword arguments are passed as-is to the subscribed handlers.

        Note that all handlers are called synchronously, in the order they subscribed, before this
        method returns.

        Args:
            topic: the topic name
        """
        for subscription in self.subscriptions[topic]:
            subscription.handler( *args, **kwargs )


    def subscribe( self, topic: str, handler: Callable ) -> Subscription:
        """
        Subscribe to the specified ``topic``

        Args:
            topic: the topic name
            handler: the handler to be called when a message is published to the topic.

        Returns:
            opaque object allowing unsubscribing
        """
        subscription = Subscription( self, topic, handler )
        self.subscriptions[topic].append( subscription )
        return subscription





from pathlib import Path
from setuptools import setup
import sys

here = Path( __file__ ).parent.resolve()
sys.path.insert( 0, here )

import mbuslite



setup(
    name = mbuslite.__title__,
    version = '0.0.0' if mbuslite.__version__ == 'git' else mbuslite.__version__,
    description = mbuslite.__description__,
    long_description = ( here / 'README.rst' ).read_text( encoding = 'utf-8' ),
    long_description_content_type = 'text/x-rst',
    url = mbuslite.__url__,
    author = mbuslite.__author__,
    author_email = mbuslite.__author_email__,
    license = mbuslite.__license__,
    packages = [ 'mbuslite' ],
    classifiers = [
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
)

